--[[

    File: fluid.lua 

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]

require 'common.glutil'
require 'common.hsv'
require 'common.pingpong'
require 'common.shader'

glush.gui.checkbox('resetField', false)			-- make this a button
glush.gui.checkbox('resetBoundary', false)		-- make this a button
glush.gui.color('paintColor')					--\ make these two radios
glush.gui.checkbox('cycleColors', true)			--/
glush.gui.checkbox('fixedHeight', true)			-- make this a radio with "fixedWidth"
glush.gui.checkbox('addToField', true)			--\
glush.gui.checkbox('addToBoundary', false)		--- make these radios
glush.gui.checkbox('removeBoundary', false)		--/
glush.gui.checkbox('showVelocityField', false)	--\ radio.  or get alpha in colors working
glush.gui.checkbox('showColorField', true)		--/
glush.gui.checkbox('showBoundary', true)
glush.gui.slider('cursorDensity', 0, 1.80882, 10)
glush.gui.slider('cursorRadius', 0, .0338251, 1)
glush.gui.slider('cursorVelScale', 0, 2.91196, 10)	-- 1.24529
glush.gui.slider('velDecay', 0, .999, 1)		-- .85 .995
glush.gui.slider('densityDecay', 0, 1, 1)	-- .995
glush.gui.slider('colorDecay', 0, 1, 1)
glush.gui.slider('timestep', 0, .025, 1)
glush.gui.slider('viscousity', 0, .00000001, 1)
glush.gui.slider('diffuseSteps', 1, 30, 50)
glush.gui.slider('gravityX', -.1,0,.1)
glush.gui.slider('gravityY', -.1, -.00005 ,.1)


-- slow copy method
fluidDim = 256
fieldFixedPointRange = 200
pingPongBackBuffers = 2		-- 2 seems more stable, 3 seemed more close to the paper
fieldTextureFormat = gl.GL_RGBA_FLOAT32
--fieldTextureFormat = gl.GL_RGB
--fieldTextureFormat = gl.GL_RGB16
--colorTextureFormat = gl.GL_RGBA_FLOAT32
colorTextureFormat = gl.GL_RGBA


vertexCode=[[
varying vec2 pos;
void main() {
	pos = gl_Vertex.xy;
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
]]

--[[
density:
	addsource
	diffuse
	advect

velocity:
	addsource
	diffuse
	project
	advect
	project
	
combined:
	addsource
	diffuse
	project (velocity only)
	advect
	project (velocity only)
jos stam does density sim then velocity sim.  think it matters that they're separate?
--]]

--[[
shaders that need set_bnd applied to them
diffuseShader u,v,density
fieldAdvectShader u,v,density
calcDivShader div,pressure
diffusePressureShader pressure
projectShader u,v
--]]


fieldClearColor={0,0,0,0}
if fieldTextureFormat == gl.GL_RGBA_FLOAT32 then
	fieldClearColor={0,0,0,0}
else
	fieldClearColor={.5,.5,.5,.5}
end

-- I'm willing to bet 'clearAll' uses 'glClearColor' which clamps to [0,1] ...
clearShader = KernelShader.new{
	code='void main() { gl_FragColor = color; }';
	uniforms={ color='vec4'; };
}


function createFluidShader(args)
	local prefix
	if fieldTextureFormat == gl.GL_RGBA_FLOAT32 then
		prefix = [[
#define FROM_CELL(v)		v
#define FROM_CELL_X(v)		v
#define FROM_CELL_Y(v)		v
#define TO_CELL(v)			v
#define FROM_DIV(v)			v
#define FROM_PRESSURE(v)	v
#define TO_DIV_PRESSURE(v)	v
]]
	else
		prefix = [[
#define MAX_FIXED				]]..fieldFixedPointRange..'.\n'..[[
#define FROM(v)					(((v) * 2. - 1.) * MAX_FIXED)
#define TO(v)					(((v / MAX_FIXED) + 1.) * .5)
#define FROM_CELL_X(v)			FROM(v)
#define FROM_CELL_Y(v)			FROM(v)
#define FROM_CELL(v)			FROM(v)
#define TO_CELL(v)				TO(v)
#define FROM_DIV(v)				FROM(v)
#define FROM_PRESSURE(v)		FROM(v)
#define TO_DIV_PRESSURE(v)		TO(v)
]]
	end
	
	return KernelShader.new{
		code=prefix..'\n'..args.code;
		texs=args.texs;
		uniforms=args.uniforms;
	}
end

-- this shader was made to be run 20x over
-- i'm adding sources in here too.  less passes.
diffuseShader = createFluidShader{
	texs={'fluidTex','fluidLastTex','sourceTex','boundaryTex'};
	uniforms={
		delta='vec2';
		alpha='float';	--set alpha to dt * fluidDim * fluidDim * viscousity
		dt='float';		--the infamous
		decay='vec4';
	};
	code=[[
void main() {
	//add source
	vec4 source = texture2D(sourceTex, pos);
	source = FROM_CELL(source);

	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	
	//diffuse
	vec4 cell = texture2D(fluidTex, pos);
	cell = FROM_CELL(cell);
	
	//dont' worry about floor()'ing pos to get the correctly aligned texel value post-interpolation
	//	on my cart texture interpolation filters don't work with float textures
	vec4 xp = texture2D(fluidLastTex, pos + dx); xp = FROM_CELL(xp);
	vec4 xn = texture2D(fluidLastTex, pos - dx); xn = FROM_CELL(xn);
	vec4 yp = texture2D(fluidLastTex, pos + dy); yp = FROM_CELL(yp);
	vec4 yn = texture2D(fluidLastTex, pos - dy); yn = FROM_CELL(yn);
	gl_FragColor = (cell + source * dt + alpha * (xp + xn + yp + yn))  / (1. + 4. * alpha);
	
	gl_FragColor *= decay;
	
#if 1
	//boundary
/** /
	vec2 boundaryGrad;
	boundaryGrad.x = texture2D(boundaryTex, pos + dy).x - texture2D(boundaryTex, pos - dy).x;
	boundaryGrad.y = texture2D(boundaryTex, pos + dx).x - texture2D(boundaryTex, pos - dx).x;
	float boundaryGradLen = length(boundaryGrad);
	boundaryGradLen = max(boundaryGradLen, .1);
	boundaryGrad *= 1. / boundaryGradLen;
	float boundaryGradDot = dot(gl_FragColor.xy, boundaryGrad);
	gl_FragColor.xy -= 1.5 * gl_FragColor.xy * boundaryGradDot;
/**/
	float boundary = texture2D(boundaryTex, pos).x;
//	boundary = step(0., boundary);
	gl_FragColor.xyz *= 1. - boundary;
#endif
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
}

--[[
field tex needs to be advected in gpu code as well as the color
however field tex needs FROM_CELL / TO_CELL and color does not
that means we need separate shaders for both

I think it might be safe to abandon the hardware advect shader,
so no point in permuting them as well
--]]
advectInSoftwareShader = createFluidShader{
	texs={'velTex', 'mediumTex'};
	uniforms={delta='vec2'; dt='float'; decay='float';};
	code=[[
void main() {
	vec4 cell = texture2D(velTex, pos);
	cell = FROM_CELL(cell);

	//advected texel coords
	//adjust it by .5 delta so that [0,0] is the center of the corner tile (rather than [.5,.5]*delta)
	vec2 atc = pos - .5 * delta - cell.xy * dt / delta;
	
	//texcoord in integer unit
	vec2 itc = atc / delta;

	//corner integer texel coordinate
	vec2 litc = floor(itc);
	vec2 uitc = litc + 1.;
	
	//fraction of interpolation along cell in S direction
	vec2 fitc = itc - litc;
	
	//make sure bilinear interpolation is disabled
	//then do the bilinear lerp in shader code
	
	vec2 ltc = litc * delta;
	vec2 utc = uitc * delta;
	
	vec4 v00 = texture2D(mediumTex, vec2(ltc.x, ltc.y)); v00 = FROM_CELL(v00);
	vec4 v01 = texture2D(mediumTex, vec2(ltc.x, utc.y)); v01 = FROM_CELL(v01);
	vec4 v10 = texture2D(mediumTex, vec2(utc.x, ltc.y)); v10 = FROM_CELL(v10);
	vec4 v11 = texture2D(mediumTex, vec2(utc.x, utc.y)); v11 = FROM_CELL(v11);
	
	vec4 v0 = mix(v00, v01, fitc.y);
	vec4 v1 = mix(v10, v11, fitc.y);
	gl_FragColor = mix(v0, v1, fitc.x);
	
	gl_FragColor *= decay;
	
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
}
advectInHardwareShader = createFluidShader{
	texs={'velTex', 'mediumTex'};
	uniforms={delta='vec2'; dt='float'; decay='float'};
	code=[[
void main() {
	vec4 cell = texture2D(velTex, pos);
	cell = FROM_CELL(cell);
	vec2 cellTc = pos - cell.xy * dt / delta;
	gl_FragColor = texture2D(mediumTex, cellTc);
	gl_FragColor *= decay;
}
]];
}

--[[ it would be ideal for the hardware to handle my bilinear interpolation
-- but it seems to be slightly skewing it in a certain corner of the screen...
if fieldTextureFormat == gl.GL_RGBA_FLOAT32 then
	fieldAdvectShader = advectInSoftwareShader
	fieldTextureFilter = gl.GL_NEAREST
else
	fieldAdvectShader = advectInHardwareShader
	fieldTextureFilter = gl.GL_LINEAR
end
if colorTextureFormat == gl.GL_RGBA_FLOAT32 then
	colorAdvectShader = advectInSoftwareShader
	colorTextureFilter = gl.GL_NEAREST
else
	colorAdvectShader = advectInHardwareShader
	colorTextureFilter = gl.GL_LINEAR
end
--]]
fieldAdvectShader = advectInHardwareShader
fieldTextureFilter = gl.GL_NEAREST
colorAdvectShader = advectInSoftwareShader
colorTextureFilter = gl.GL_NEAREST

calcDivShader = createFluidShader{
	texs={'fluidTex'};
	uniforms={delta='vec2'};
	code=[[
void main() {
	float h = delta.x;	//1/N in stam's paper, and N is the x and y dimensions
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float uxp = texture2D(fluidTex, pos + dx).x; uxp = FROM_CELL_X(uxp);
	float uxn = texture2D(fluidTex, pos - dx).x; uxn = FROM_CELL_X(uxn);
	float vyp = texture2D(fluidTex, pos + dy).y; vyp = FROM_CELL_Y(vyp);
	float vyn = texture2D(fluidTex, pos - dy).y; vyn = FROM_CELL_Y(vyn);
	vec2 uvGrad = vec2(uxp-uxn, vyp-vyn) * delta;
	float div = -.5 * (uvGrad.x + uvGrad.y);
	//div, pressure, 0, 0
	gl_FragColor = vec4(div, 0., 0., 0.);
	gl_FragColor = TO_DIV_PRESSURE(gl_FragColor);
}
]];
}

diffusePressureShader = createFluidShader{
	texs={'divTex', 'divLastTex'};
	uniforms={delta='vec2'};
	code=[[
void main() {
	float divLast = texture2D(divTex, pos).x;
	divLast = FROM_DIV(divLast);
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float pxp = texture2D(divLastTex, pos + dx).y; pxp = FROM_PRESSURE(pxp);
	float pxn = texture2D(divLastTex, pos - dx).y; pxn = FROM_PRESSURE(pxn);
	float pyp = texture2D(divLastTex, pos + dy).y; pyp = FROM_PRESSURE(pyp);
	float pyn = texture2D(divLastTex, pos - dy).y; pyn = FROM_PRESSURE(pyn);
	float pressure = (divLast + pxp + pxn + pyp + pyn) * .25;
	gl_FragColor = vec4(
		divLast,	//maintain div
		pressure,	//diffuse pressure,
		0., 0.);
	gl_FragColor = TO_DIV_PRESSURE(gl_FragColor);
}
]];
}

projectShader = createFluidShader{
	texs={'fluidTex', 'divTex'};
	uniforms={delta='vec2'};
	code=[[
void main() {
	vec4 cell = texture2D(fluidTex, pos); cell = FROM_CELL(cell);
	vec2 dx = vec2(delta.x, 0.);
	vec2 dy = vec2(0., delta.y);
	float pxp = texture2D(divTex, pos + dx).y; pxp = FROM_PRESSURE(pxp);
	float pxn = texture2D(divTex, pos - dx).y; pxn = FROM_PRESSURE(pxn);
	float pyp = texture2D(divTex, pos + dy).y; pyp = FROM_PRESSURE(pyp);
	float pyn = texture2D(divTex, pos - dy).y; pyn = FROM_PRESSURE(pyn);
	vec2 pGrad = vec2(pxp-pxn, pyp-pyn) / delta;
	gl_FragColor = vec4(cell.xy - .5 * pGrad, cell.zw);
	gl_FragColor = TO_CELL(gl_FragColor);
}
]];
}

displayShader = createFluidShader{
	texs={'fluidTex', 'gradTex'};
	code=[[
#define pi ]]..math.pi..'\n'..[[
#define maxSpeed .0001
#define maxDensity .01

vec3 hsv2rgb(vec3 hsv) {
	vec3 hue = texture2D(gradTex, vec2(hsv[0], .5)).xyz;
	vec3 lum = vec3(hsv[2]);
	return mix(lum, hue, hsv[1]);
	
}

void main() {
	vec4 cell = texture2D(fluidTex, pos); cell = FROM_CELL(cell);
#if 0	//velocity in there too ...
	float speed = length(cell.xy);
	cell.xy /= max(speed, maxSpeed);
	cell.xy *= .49;	//[-1,1] -> [0,1]
	cell.xy += .5;
	cell.z /= maxDensity;
	cell.z = fract(cell.z);
	gl_FragColor = vec4(cell.xzy, 1.);
#endif
#if 0	//heat based on density
	gl_FragColor = texture2D(gradTex, vec2(cell.z / maxDensity, .5));
#endif
#if 1
	float speed = sqrt(cell.x * cell.x + cell.y * cell.y);
	float angle = atan(cell.y, cell.x);
	vec3 c = vec3(
		fract((angle / (2. * pi)) + .5),
		1. - fract(speed / maxSpeed),
		fract(cell.z / maxDensity));
	c = hsv2rgb(c);
	gl_FragColor = vec4(c.xyz, 1.);
#endif
}
]];
}

-- usually the color field is drawn without a shader
-- and bilinear filtering smoothes over the cell tiles
-- but this isn't available in hardware for float textures...
--
-- this might be a ripoff of advectInSoftwareShader
-- which makes me want to consolidate code...
displayColorLerpShader = createFluidShader{
	texs={'colorTex'};
	uniforms={delta='vec2'};
	code=[[
void main() {
	vec2 itc = pos / delta;
	vec2 litc = floor(itc);
	vec2 uitc = litc + 1.;
	vec2 fitc = itc - litc;
	vec2 ltc = litc * delta;
	vec2 utc = uitc * delta;
	vec4 v00 = texture2D(colorTex, vec2(ltc.x, ltc.y));
	vec4 v01 = texture2D(colorTex, vec2(ltc.x, utc.y));
	vec4 v10 = texture2D(colorTex, vec2(utc.x, ltc.y));
	vec4 v11 = texture2D(colorTex, vec2(utc.x, utc.y));
	vec4 v0 = mix(v00, v01, fitc.y);
	vec4 v1 = mix(v10, v11, fitc.y);
	gl_FragColor = mix(v0, v1, fitc.x);
}
]];
}

displayFieldShader = createFluidShader{
	texs={'fieldTex'};
	uniforms={delta='vec2'};
	code=[[
void main() {
	vec2 itc = pos / delta;
	vec2 litc = floor(itc);
	vec2 uitc = litc + 1.;
	vec2 fitc = itc - litc;
	vec2 ltc = litc * delta;
	vec2 utc = uitc * delta;
	vec4 v00 = texture2D(fieldTex, vec2(ltc.x, ltc.y));
	vec4 v01 = texture2D(fieldTex, vec2(ltc.x, utc.y));
	vec4 v10 = texture2D(fieldTex, vec2(utc.x, ltc.y));
	vec4 v11 = texture2D(fieldTex, vec2(utc.x, utc.y));
	vec4 v0 = mix(v00, v01, fitc.y);
	vec4 v1 = mix(v10, v11, fitc.y);
	vec4 field = mix(v0, v1, fitc.x);
	field = FROM_CELL(field);
	
	float contour = length(field.xy);
	contour *= 1000.;
	contour = 1. / (contour + .33);
	contour = fract(contour);
	const float stepRatio = .02;	//how much is stepping up and down
	const float bandRatio = .02;	//how much is levelled off
	contour = smoothstep(.5, .5 + stepRatio, contour)
		- smoothstep(.5 + stepRatio + bandRatio, .5 + 2. * stepRatio + bandRatio, contour);
	
	vec3 color = vec3(field.xy, 0.);
	color = abs(color);
	color *= 250.;	//field velocity -> color range
	color = mix(color, vec3(1.), contour);
	gl_FragColor = vec4(color.xyz, 1.);
}
]];
}

addCursorVelDensityShader = createFluidShader{
	texs={'fluidTex'};
	uniforms={sourcePoint='vec2'; sourceVel='vec2'; sourceDensity='float'; sourceRadius='float'};
	code=[[
void main() {
	vec4 cell = texture2D(fluidTex, pos);
	cell = FROM_CELL(cell);
	
	vec2 del = pos - sourcePoint;
	float lenSq = dot(del,del);
	if (lenSq < sourceRadius*sourceRadius) {
		float unitlen = sqrt(lenSq) / sourceRadius;
		float infl = sqrt(1. - unitlen);
		cell.xyz += infl * vec3(sourceVel, sourceDensity);
	}
	
	gl_FragColor = TO_CELL(cell);
}
]];
}

addCursorColorShader = createFluidShader{
	texs={'colorTex'};
	uniforms={sourcePoint='vec2'; sourceRadius='float'; sourceColor='vec3'};
	code=[[
void main() {
	vec4 cell = texture2D(colorTex, pos);
	vec2 del = pos - sourcePoint;
	float lenSq = dot(del,del);
	if (lenSq < sourceRadius*sourceRadius) {
		float unitlen = sqrt(lenSq) / sourceRadius;
		float infl = sqrt(1. - unitlen);
		vec4 sourceColor4 = vec4(sourceColor, 1.);
		cell = mix(cell, sourceColor4, infl);
	}
	gl_FragColor = cell;
}
]];
}

gl.glUseProgramObject(0)

fluidDimViewport = {0, 0, fluidDim, fluidDim}
fluidDelta = {1/fluidDim, 1/fluidDim}		-- hardcode in shaders?

-- rgba: u,v, density
fluidTexs = PingPong.new{
	numBuffers=pingPongBackBuffers;
	width=fluidDim;
	height=fluidDim;
	internalFormat=fieldTextureFormat;
	minFilter = fieldTextureFilter;
	magFilter = fieldTextureFilter;
}

-- rgba: color
colorTexs = PingPong.new{
	width=fluidDim;
	height=fluidDim;
	internalFormat=colorTextureFormat;
	minFilter=colorTextureFilter;
	magFilter=colorTextureFilter;
}

-- rgba: source u, v, density
sourceTex = PingPong.new{
	numBuffers=1;
	width=fluidDim;
	height=fluidDim;
	internalFormat=fieldTextureFormat;
	data=function(u,v) return 0,0,0,0 end;
	minFilter=fieldTextureFilter;
	magFilter=fieldTextureFilter;
}

-- rgba: solid, 0,0,0
boundaryTexs = PingPong.new{
	width=fluidDim;
	height=fluidDim;
	internalFormat=colorTextureFormat;
	data=function(v,y) return 0,0,0,0 end;
	minFilter=colorTextureFilter;
	magFilter=colorTextureFilter;
	--wrap={s=gl.GL_NEAREST, t=gl.GL_NEAREST};	-- gives a glreport error once, but works better...
}

-- rgba: divergence, pressure
divTexs = PingPong.new{
	numBuffers=pingPongBackBuffers;
	width=fluidDim;
	height=fluidDim;
	internalFormat=fieldTextureFormat;
	minFilter=fieldTextureFilter;
	magFilter=fieldTextureFilter;
}

function clearAll(pingpong, color)
--[[ glClearColor based
	pingpong:clearAll(color)
--]]
-- [[ shader based (doesn't get clamped)
	for i=1,#pingpong.hist do
		pingpong:swap()
		pingpong:draw{
			shader=clearShader,
			uniforms={
				color=color
			}
		}
	end
--]]
end

function doResetField()
	clearAll(fluidTexs, fieldClearColor)
	clearAll(divTexs, fieldClearColor)
	clearAll(colorTexs, {0,0,0,0})
	clearAll(sourceTex, {gravityX, gravityY, 0, 0})
end

function doResetBoundary()
	clearAll(boundaryTexs, {0,0,0,0})
end

function diffuse()
	for k=1,diffuseSteps do
		glreport('before diffuse')
		fluidTexs:swap()
		fluidTexs:draw{
			shader=diffuseShader;
			texs={fluidTexs:prev(), fluidTexs:last(), sourceTex:cur(), boundaryTexs:cur()};
			uniforms={
				delta = fluidDelta;
				alpha = timestep * fluidDim * fluidDim * viscousity;
				dt = timestep;
				decay = {velDecay, velDecay, densityDecay, 0.};
			};
		}
		glreport('after diffuse')
		-- set_bnd u, v, density
	end
end

function advectColor()
	glreport('before advect color')
	colorTexs:swap()
	colorTexs:draw{
		shader=colorAdvectShader;
		texs={fluidTexs:prev(), colorTexs:prev()};
		uniforms={
			delta=fluidDelta;
			dt=timestep;
			--[[
			stam's paper does decay in the diffuse() step
			but i never write to the color buffer per-frame anywhere else
			so i figured i'd incorperate it here.
			of coures this uses the same advect shader as the density and velocity below
			so it'll have to set a default '1' for its updates
			--]]
			decay=colorDecay;
		};
	}
	glreport('after advect color')
end

function advect()
	glreport('before advect')
	fluidTexs:swap()
	fluidTexs:draw{
		shader=fieldAdvectShader;
		texs={fluidTexs:prev(), fluidTexs:last()};
		uniforms={
			delta=fluidDelta;
			dt=timestep;
			decay=1;
		};
	}
	glreport('after advect')
	-- set_bnd u, v, density
end

function project()
	-- 1) calc div and clear p (in an aux tex somewhere)
	glreport('before calcDiv')
	divTexs:draw{
		shader=calcDivShader;
		texs={fluidTexs:cur()};
		uniforms={
			delta=fluidDelta;
		};
	}
	glreport('after calcDiv')
	-- set_bnd div, pressure (even though pressure is only being cleared, eh?)	

	-- 2) diffuse p ... strangely... perpetually treating div as the source
	for k=1,diffuseSteps do
		glreport('before diffuse pressure')
		divTexs:swap()
		divTexs:draw{
			shader=diffusePressureShader;
			texs={divTexs:prev(), divTexs:last()};
			uniforms={
				delta=fluidDelta;
			};
		}
		glreport('after diffuse pressure')
		-- set_bnd pressure
	end
	-- ... that means converge it until it makes up for the difference of divergences (the pressure!)
	-- 3) 
	glreport('before project')
	fluidTexs:swap()
	fluidTexs:draw{
		shader=projectShader;
		texs={fluidTexs:prev(), divTexs:cur()};
		uniforms={
			delta=fluidDelta;
		};
	}
	glreport('after project')
	-- set_bnd u, v
end

function iterate()
	diffuse()
	project()
	advectColor()
	advect()
	project()
end

function addMouseSource(pos,vel,color)
	if addToField then
		fluidTexs:swap()
		fluidTexs:draw{
			texs={fluidTexs:prev()};
			shader=addCursorVelDensityShader;
			uniforms={
				sourcePoint=pos;
				sourceVel=vel;
				sourceDensity=cursorDensity;
				sourceRadius=cursorRadius;
			};
		}
		colorTexs:swap()
		colorTexs:draw{
			texs={colorTexs:prev()};
			shader=addCursorColorShader;
			uniforms={
				sourcePoint=pos;
				sourceColor=color;
				sourceRadius=cursorRadius;
			};
		}
	end
	if addToBoundary or removeBoundary then
		local color = {0,0,0}
		if addToBoundary then color = {1,1,1} end
		boundaryTexs:swap()
		boundaryTexs:draw{
			texs={boundaryTexs:prev()};
			shader=addCursorColorShader;
			uniforms={
				sourcePoint=pos;
				sourceColor=color;
				sourceRadius=cursorRadius;
			};
		}
	end
end

gl.glDisable(gl.GL_DEPTH_TEST)

-- initiali resets
resetField = true
resetBoundary = true

local lastFieldMousePos = {0,0}
function glush.display()
	glreport('begin display')

	local viewport = gl.glGet(gl.GL_VIEWPORT)
	local aspectRatio
	if fixedHeight then
		aspectRatio = viewport[3] / viewport[4]
	else
		aspectRatio = viewport[4] / viewport[3]
	end
	
	-- set up for viewport fbo once
	-- assert that all your fbo targets are the same size
	gl.glViewport(0,0,fluidDim,fluidDim)
	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glLoadIdentity()
	gl.glOrtho(0,1,0,1,-1,1)
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glLoadIdentity()
	
	if resetField then doResetField() resetField = false end
	if resetBoundary then doResetBoundary() resetBoundary = false end

	local mouseX,mouseY,mouseButton = glush.mouse()
	local fieldMousePos
	if fixedHeight then
		fieldMousePos = { mouseX*aspectRatio - (aspectRatio-1)/2, mouseY }
	else
		fieldMousePos = { mouseX, mouseY*aspectRatio - (aspectRatio-1)/2 }
	end
	local fieldMouseVel = {
		(fieldMousePos[1] - lastFieldMousePos[1]) * cursorVelScale,
		(fieldMousePos[2] - lastFieldMousePos[2]) * cursorVelScale
	}
	local fieldMouseColor = paintColor
	if cycleColors then fieldMouseColor = hsv2rgb{glush.timer()/5-.3, math.sin(glush.timer()/3)*.5+.5, 1} end
	if mouseButton ~= 0 then
		addMouseSource(fieldMousePos, fieldMouseVel, fieldMouseColor)
	end
	
	lastFieldMousePos[1] = fieldMousePos[1]
	lastFieldMousePos[2] = fieldMousePos[2]

	iterate()

	-- set up for rendering
	gl.glViewport(unpack(viewport))
	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glLoadIdentity()
	if fixedHeight then
		gl.glOrtho(-(aspectRatio-1)/2, 1+(aspectRatio-1)/2, 0, 1, -1, 1)
	else
		gl.glOrtho(0, 1, -(aspectRatio-1)/2, 1+(aspectRatio-1)/2, -1, 1)
	end
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glLoadIdentity()


	-- here's the colorfield display

	gl.glColor(1,1,1)
	gl.glEnable(gl.GL_TEXTURE_2D)
	
	if showVelocityField then
		-- rip off of what you see at this site:
		-- red = x velocity
		-- green = y velocity
		-- and white contour lines of the velocity magnitude
		gl.glUseProgramObject(displayFieldShader)
		gl.glUniformf('delta', fluidDelta)
		local fluidTex = fluidTexs:cur()
		fluidTex:bind()
		drawScreenQuad()
		fluidTex:unbind()
		gl.glUseProgramObject(0)
	end
	
	if showColorField then
		-- if using float textures (without hardware bilienar interpolation) 
		-- then use a shader to do the visual interpolation
		-- if not using float textures then we just switch on hardware bilinear interpolation
		if colorTextureFormat == gl.GL_RGBA_FLOAT32 then
			gl.glUseProgramObject(displayColorLerpShader)
			gl.glUniformf('delta', fluidDelta)
		end
		local colorTex = colorTexs:cur()
		colorTex:bind()
		-- switch to linear filter for pretty drawing
		gl.glTexParameter(colorTex.target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
		drawScreenQuad()
		-- switch back for subsequent shader operations
		gl.glTexParameter(colorTex.target, gl.GL_TEXTURE_MAG_FILTER, colorTextureFilter)
		colorTex:unbind()
	end

	if showBoundary then
		gl.glBlendFunc(gl.GL_SRC_COLOR, gl.GL_ONE_MINUS_SRC_COLOR)
		gl.glEnable(gl.GL_BLEND)
		local boundaryTex = boundaryTexs:cur()
		boundaryTex:bind()
		gl.glTexParameter(boundaryTex.target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
		drawScreenQuad()
		gl.glTexParameter(boundaryTex.target, gl.GL_TEXTURE_MAG_FILTER, fieldTextureFilter)
		boundaryTex:unbind()
		gl.glDisable(gl.GL_BLEND)
	end
	if colorTextureFormat == gl.GL_RGBA_FLOAT32 then
		gl.glUseProgramObject(0)
	end
	
	gl.glDisable(gl.GL_TEXTURE_2D)

	glreport('end display')
end
